################################# LICENSE ##################################
# Copyright (c) 2009, South African Astronomical Observatory (SAAO)        #
# All rights reserved.                                                     #
#                                                                          #
# Redistribution and use in source and binary forms, with or without       #
# modification, are permitted provided that the following conditions       #
# are met:                                                                 #
#                                                                          #
#     * Redistributions of source code must retain the above copyright     #
#       notice, this list of conditions and the following disclaimer.      #
#     * Redistributions in binary form must reproduce the above copyright  #
#       notice, this list of conditions and the following disclaimer       #
#       in the documentation and/or other materials provided with the      #
#       distribution.                                                      #
#     * Neither the name of the South African Astronomical Observatory     #
#       (SAAO) nor the names of its contributors may be used to endorse    #
#       or promote products derived from this software without specific    #
#       prior written permission.                                          #
#                                                                          #
# THIS SOFTWARE IS PROVIDED BY THE SAAO ''AS IS'' AND ANY EXPRESS OR       #
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           #
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE   #
# DISCLAIMED. IN NO EVENT SHALL THE SAAO BE LIABLE FOR ANY                 #
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL       #
# DAMAGES (INCte: 2007/05/26
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)    #
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,      #
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          #
# POSSIBILITY OF SUCH DAMAGE.                                              #
############################################################################


#!/usr/bin/env python

"""
FINDCAL--From the list of observations for a night,
find the different settups that will be required

Author                 Version      Date
-----------------------------------------------
S M Crawford (SAAO)    0.1          19 Jun 2011

"""

from pyraf import iraf
from pyraf.iraf import pysalt
import saltsafemysql as saltmysql


def findcal(obsdate, sdbhost, sdbname, sdbuser, password):
    """Find all of the unique configuration settings used on a night 
       and insert it into the database
  
       obsdate--Observation date in YYYYMMDD

       sdbhost--host name for sdb

       sdbname--name of the sdb

       sdbuser--user for the sdb

       password--sdb password
    """

    #connect to the db
    sdb=saltmysql.connectdb(sdbhost, sdbname, sdbuser, password)

    #get the nightinfo id
    logic="Date='%4s-%2s-%2s'" % (obsdate[0:4], obsdate[4:6], obsdate[6:8])
    results=saltmysql.select(sdb, 'NightInfo_Id', 'NightInfo', logic)
    night_id=results[0][0]

    #select all the scam data from this obsdate
    cmd_select='FileName, FileData_Id, CCDTYPE, DETMODE, CCDSUM, GAINSET, ROSPEED, FILTER'
    cmd_table=''' FileData 
  left join FitsHeaderImage using (FileData_Id) 
  left join FitsHeaderSalticam using (FileData_Id)
'''
    cmd_logic="FileName like 'S" + obsdate+"%' and CCDTYPE='OBJECT'"

    results=saltmysql.select(sdb, cmd_select, cmd_table, cmd_logic)

    #loop through all the results and return only the Set of identical results
    caldict=create_caldict(results)
               
    #insert the results into the database
    for k in caldict:
        #first check to see if it has already been entered
        record=saltmysql.select(sdb, 'FileData_Id', 'SalticamNightlyCalibration', 'FileData_Id=%i' % k)
        print record
        if len(record)<1:
           #insert cal flat
           cmd_insert='NightInfo_Id=%i, FileData_Id=%i, CalibrationType_Id=1' % (night_id, k)
           saltmysql.insert(sdb, cmd_insert, 'SalticamNightlyCalibration')
           #insert sky flat
           cmd_insert='NightInfo_Id=%i, FileData_Id=%i, CalibrationType_Id=2' % (night_id, k)
           print cmd_insert
           saltmysql.insert(sdb, cmd_insert, 'SalticamNightlyCalibration')
        print k, " ".join([str(k) for k in caldict[k]])


    #select all the RSS data from this obsdate
    rssheaderlist='CCDTYPE, DETMODE, OBSMODE, CCDSUM, GAINSET, ROSPEED, FILTER, GRATING, GR_STA, AR_STA, MASKID'
    cmd_select='FileName,FileData_Id,  CCDTYPE, DETMODE, OBSMODE, CCDSUM, GAINSET, ROSPEED, FILTER, GRATING, GR_STA, AR_STA, MASKID'
    cmd_table=''' FileData 
  left join FitsHeaderImage using (FileData_Id) 
  left join FitsHeaderRss using (FileData_Id)
  join ProposalCode using (ProposalCode_Id)
'''
    cmd_logic="FileName like 'P" + obsdate+"%' and CCDTYPE='OBJECT' and Proposal_Code not like 'CAL_SPST'"

    results=saltmysql.select(sdb, cmd_select, cmd_table, cmd_logic)
    print results

    #loop through all the results and return only the Set of identical results
    caldict=create_caldict(results)

    #insert the scam results into the database
    for k in caldict:
        #first check to see if it has already been entered
        record=saltmysql.select(sdb, 'FileData_Id', 'RssNightlyCalibration', 'FileData_Id=%i' % k)
       
        if len(record)<1:

           #insert bias information
           cmd_insert='NightInfo_Id=%i, FileData_Id=%i, CalibrationType_Id=4' % (night_id, k)
           saltmysql.insert(sdb, cmd_insert, 'RssNightlyCalibration')

           #insert sky flat
           cmd_insert='NightInfo_Id=%i, FileData_Id=%i, CalibrationType_Id=1' % (night_id, k)
           saltmysql.insert(sdb, cmd_insert, 'RssNightlyCalibration')
           #insert spectral flat
           if caldict[k][2]=='SPECTROSCOPY':
               cmd_insert='NightInfo_Id=%i, FileData_Id=%i, CalibrationType_Id=3' % (night_id, k)
               saltmysql.insert(sdb, cmd_insert, 'RssNightlyCalibration')
               #insert sky flat

               #Add a check to make sure that something similar to this one isn't already in there
               
               cmd_insert='NightInfo_Id=%i, FileData_Id=%i, CalibrationType_Id=5' % (night_id, k)
               print cmd_insert
               saltmysql.insert(sdb, cmd_insert, 'RssNightlyCalibration')
        print k," ".join([str(k) for k in caldict[k]])

    #check scam data

    #check rss data
    #checkfortaken(sdb, 'RssNightlyCalibration', rssheaderlist)

    return

def checkfortaken(sdb, table, plist):
    """Check the sdb for any calibration taken--if it is taken, then fill in the table appropriately

       table: table to check
       plist: header parameters to compare
    """
    
    #get all the data not taken
    record=saltmysql.select(sdb, 'FileData_Id, NightInfo_id, Calibration_Type', 
          '%s join CalibrationType using (CalibrationType_Id)' % table, 'CalibrationTaken is NULL')
    if len(record)<1: return

    for fid, nid, ctype in record:
        #for a given file data, check to see if any calibration frames were
        print fid, nid, ctype

        #now get the parameters of the data
    
        #now get a list of all data taken in the last three days with those parameters

        

def create_caldict(results):
    """Sort through the results from the query and create the calibration dictionary

    """
        #loop through all the results and return only the Set of identical results
    caldict={}
    for r in results:
        clist=r[2:]
        for k in caldict:
           if compare_configs(clist, caldict[k]):
                 clist=[]
        if clist: caldict[r[1]]=clist
    return caldict

def compare_configs(c1, c2):
    """Compare two configurations list to see if they are the same.  
       returns true if they are the same
    """
    for i in range(len(c1)):
        if c1[i]!=c2[i]: return False
    return True


if __name__=='__main__':
   import sys
   findcal(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4], sys.argv[5])
